import 'dart:async';

import 'package:dart_appwrite/dart_appwrite.dart';
import 'package:dart_appwrite/models.dart' as appwrite_models;
import 'package:lib/src/interface/response_model.dart';
import 'package:lib/src/request_response/db_request.dart';
import 'package:lib/src/request_response/db_response.dart';
import 'package:lib/src/exceptions/id_null.dart';
import 'package:lib/src/interface/client_function.dart';
import 'package:lib/src/interface/db_model.dart';

abstract class BaseDatabase<T extends DbModel, T1 extends ResponseModel> extends ClientFunction {
  static const String itemDataField = "data";
  final Future<T1> Function(Map<String, dynamic>) itemFromDbMap;
  late final Databases databases;

  String get databaseId;
  String get collectionId;

  BaseDatabase({required super.host, required super.projectId, required super.apiKey, required this.itemFromDbMap}) {
    databases = Databases(client);
  }

  Future<DbResponse<T1>> create(DbRequest<T> request) async {
    var id = request.item.id ?? ID.unique();
    var document = await databases.createDocument(databaseId: databaseId, collectionId: collectionId, documentId: id, data: request.item.toDbData(), permissions: request.item.permissions?.dbPermissions);
    return _responseFromDocument(document);
  }

  Future<DbResponse<T1>> get(DbRequest<String> request) async {
    var document = await databases.getDocument(databaseId: databaseId, collectionId: collectionId, documentId: request.item);
    return _responseFromDocument(document);
  }

  Stream<DbResponse<T1>> getFromQueries(DbRequest<Iterable<String>> request) async* {
    var queryResult = await databases.listDocuments(databaseId: databaseId, collectionId: collectionId, queries: request.item.toList());

    for (var document in queryResult.documents) {
      yield await _responseFromDocument(document);
    }
  }

  Stream<DbResponse<T1>> getAll(DbRequest request) async* {
    var queryResults = getFromQueries(DbRequest(item: [], requestingUser: request.requestingUser));
    await for (var t in queryResults) {
      yield t;
    }
  }

  Future<DbResponse<T1>> update(DbRequest<T> request) async {
    var id = request.item.id;
    if (id == null) {
      throw IdNullException(request.item);
    }
    var document = await databases.updateDocument(databaseId: databaseId, collectionId: collectionId, documentId: id, data: request.item.toDbData(), permissions: request.item.permissions?.dbPermissions);
    return _responseFromDocument(document);
  }

  Future<DbResponse<T1>> delete(DbRequest<String> request) async {
    var itemToDelete = await get(request);
    await databases.deleteDocument(databaseId: databaseId, collectionId: collectionId, documentId: request.item);
    return itemToDelete;
  }

  Future<DbResponse<T1>> _responseFromDocument(appwrite_models.Document document) async {
    return DbResponse(item: await itemFromDbMap(document.toMap()[itemDataField]));
  }
}
